diff --git a/gdk/quartz/gdkscreen-quartz.c b/gdk/quartz/gdkscreen-quartz.c
index 219338832e..76bd1205e4 100644
--- a/gdk/quartz/gdkscreen-quartz.c
+++ b/gdk/quartz/gdkscreen-quartz.c
@@ -68,7 +68,10 @@ static void display_reconfiguration_callback (CGDirectDisplayID            displ
                                               CGDisplayChangeSummaryFlags  flags,
                                               void                        *userInfo);
 static const double dpi = 96.0;
-static gint get_mm_from_pixels (NSScreen *screen, int pixels);
+static gint get_mm_width_from_pixels  (NSScreen *screen,
+                                       int       pixels);
+static gint get_mm_height_from_pixels (NSScreen *screen,
+                                       int       pixels);
 
 G_DEFINE_TYPE (GdkQuartzScreen, gdk_quartz_screen, GDK_TYPE_SCREEN);
 
@@ -189,8 +192,8 @@ gdk_quartz_screen_calculate_layout (GdkQuartzScreen *screen)
         monitor->scale_factor = [(id <ScaleFactor>) nsscreen backingScaleFactor];
       else
         monitor->scale_factor = 1;
-      monitor->width_mm = get_mm_from_pixels(nsscreen, monitor->geometry.width);
-      monitor->height_mm = get_mm_from_pixels(nsscreen, monitor->geometry.height);
+      monitor->width_mm = get_mm_width_from_pixels(nsscreen, monitor->geometry.width);
+      monitor->height_mm = get_mm_height_from_pixels(nsscreen, monitor->geometry.height);
       monitor->refresh_rate = 0; // unknown
       monitor->manufacturer = NULL; // unknown
       monitor->model = NULL; // unknown
@@ -339,10 +342,51 @@ gdk_quartz_screen_get_height (GdkScreen *screen)
 }
 
 static gint
-get_mm_from_pixels (NSScreen *screen, int pixels)
+get_mm_width_from_pixels (NSScreen *screen,
+                          int       pixels)
 {
-  const float mm_per_inch = 25.4;
-  return (pixels / dpi) * mm_per_inch;
+  NSDictionary *description;
+  NSSize        displayPixelSize;
+  CGSize        displayPhysicalSize;
+  gdouble       ppmm;
+
+  GDK_QUARTZ_ALLOC_POOL;
+
+  description = [screen deviceDescription];
+  displayPixelSize = [[description objectForKey:NSDeviceSize] sizeValue];
+  displayPhysicalSize =
+    CGDisplayScreenSize([[description objectForKey:@"NSScreenNumber"]
+                         unsignedIntValue]);
+
+  ppmm = (gdouble) displayPixelSize.width / (geouble) displayPhysicalSize.width;
+
+  GDK_QUARTZ_RELEASE_POOL;
+
+  return pixels / ppmm;
+}
+
+static gint
+get_mm_height_from_pixels (NSScreen *screen,
+                           int       pixels)
+{
+  NSDictionary *description;
+  NSSize        displayPixelSize;
+  CGSize        displayPhysicalSize;
+  gdouble       ppmm;
+
+  GDK_QUARTZ_ALLOC_POOL;
+
+  description = [screen deviceDescription];
+  displayPixelSize = [[description objectForKey:NSDeviceSize] sizeValue];
+  displayPhysicalSize =
+    CGDisplayScreenSize([[description objectForKey:@"NSScreenNumber"]
+                         unsignedIntValue]);
+
+  ppmm = (gdouble) displayPixelSize.height / (gdouble) displayPhysicalSize.height;
+
+  GDK_QUARTZ_RELEASE_POOL;
+
+  return pixels / ppmm;
 }
 
 static gchar *
@@ -388,15 +432,15 @@ get_nsscreen_for_monitor (gint monitor_num)
 static gint
 gdk_quartz_screen_get_width_mm (GdkScreen *screen)
 {
-  return get_mm_from_pixels (get_nsscreen_for_monitor (0),
-                             GDK_QUARTZ_SCREEN (screen)->width);
+  return get_mm_width_from_pixels (get_nsscreen_for_monitor (0),
+                                   GDK_QUARTZ_SCREEN (screen)->width);
 }
 
 static gint
 gdk_quartz_screen_get_height_mm (GdkScreen *screen)
 {
-  return get_mm_from_pixels (get_nsscreen_for_monitor (0),
-                             GDK_QUARTZ_SCREEN (screen)->height);
+  return get_mm_height_from_pixels (get_nsscreen_for_monitor (0),
+                                    GDK_QUARTZ_SCREEN (screen)->height);
 }
 
 static void
